/* 

LBMix - A mixer for Crystal Semiconductor IOCTL90 by Lesha Bogdanow
Copyright (C) 1999-2000  Lesha Bogdanow

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include "LBMix.h"
#include "ioctl90.h"
#include "mixerapi.h"
#include "PipeMix.h"

#define PIPE_WAIT 200		// Wait for pipe (ms)
#define DEF_INFO_BUF_SIZE 1024	// Size for info buffer if driver does not return info

// Additional flags for mixerapi
#define FLAG_MUTE 1
#define FLAG_RELEASE 2

// User window messages parameters
#define USER_INIT 0
#define USER_QUERY 1


HAB   hab;			    /* anchor block for the process */
HMQ   hmq;		    /* handle to the process' message queue */
HWND  hwndMain;			/* handle to the main client window */
HSWITCH hswitch;
HEV	hev=0;			//My semaphore


// Definitions for Controls structure
#define FLAG_NONE	0	// Dummy
#define FLAG_ONLYLEFT	1	// Only left control is used
#define FLAG_STEREO	2	// Both is valid. 
#define FLAG_MUTABLE	4	// Mute is valid
#define FLAG_LOCKABLE	8	// Can be locked, lock is valid
#define FLAG_MULTIPLE	0x10	// This input selector supports multiple selections
#define FLAG_ONLYRIGHT	0x20	// Only right control is used

#define NO_INPUT_SEL    0	/* Value to place into .sel field when control
				   does not appear in input selector */
enum CtlTypes{
   TYPE_DEFAULT,
   TYPE_SELECTOR,
   TYPE_3D,
   TYPE_TONE};

struct t_Controls{
   int setfunc;		// IOCTL set function no
   int qfunc;		// IOCTL query function no
   int flags;		// Flags, see above
   enum CtlTypes type;	// Control type
   int sel;		// Value representing the control in input selector or NO_INPUT_SEL
   int groupID,leftID,rightID,muteID,bothID,lockID;	// Window controls IDs
   int ldescID,rdescID;					// ...continued
   int activeID,nameID;					// Settings controls IDs
   unsigned char defname[32];				// defaut name
   int supported;					// TRUE if supported
   int active;						// TRUE if active
   unsigned char name[32];				// name
   int left,right,mute,both,lock;			// Current values
   } Controls[]={ 
{0x40,0x60,FLAG_ONLYLEFT|FLAG_MUTABLE,TYPE_DEFAULT,NO_INPUT_SEL,	// MonoIn
IDC_MONOBOX,IDC_MONOLEFT,IDC_MONORIGHT,IDC_MONOMUTE,IDC_MONOBOTH,IDC_MONOLOCK,
0,0,
IDC_MONOACT,IDC_MONONAME,
"MonoIn",
1,1,"MonoIn",100,100,0,0,0},
{0x41,0x61,FLAG_STEREO|FLAG_MUTABLE,TYPE_DEFAULT,I90SRC_PHONE,	// Phone
IDC_PHONEBOX,IDC_PHONELEFT,IDC_PHONERIGHT,IDC_PHONEMUTE,IDC_PHONEBOTH,IDC_PHONELOCK,
0,0,
IDC_PHONEACT,IDC_PHONENAME,
"Phone",
1,1,"Phone",100,100,0,0,0},
{0x42,0x62,FLAG_STEREO|FLAG_MUTABLE,TYPE_DEFAULT,I90SRC_MIC,		// Mic
IDC_MICBOX,IDC_MICLEFT,IDC_MICRIGHT,IDC_MICMUTE,IDC_MICBOTH,IDC_MICLOCK,
0,0,
IDC_MICACT,IDC_MICNAME,
"Mic",
1,1,"Mic",100,100,0,0,0},
{0x43,0x63,FLAG_STEREO|FLAG_MUTABLE,TYPE_DEFAULT,I90SRC_LINE,	// Line
IDC_LINEBOX,IDC_LINELEFT,IDC_LINERIGHT,IDC_LINEMUTE,IDC_LINEBOTH,IDC_LINELOCK,
0,0,
IDC_LINEACT,IDC_LINENAME,
"Line",
1,1,"Line",100,100,0,0,0},
{0x44,0x64,FLAG_STEREO|FLAG_MUTABLE,TYPE_DEFAULT,I90SRC_CD,		// CD
IDC_CDBOX,IDC_CDLEFT,IDC_CDRIGHT,IDC_CDMUTE,IDC_CDBOTH,IDC_CDLOCK,
0,0,
IDC_CDACT,IDC_CDNAME,
"CD",
1,1,"CD",100,100,0,0,0},
{0x45,0x65,FLAG_STEREO|FLAG_MUTABLE,TYPE_DEFAULT,I90SRC_VIDEO,	// Video
IDC_VIDEOBOX,IDC_VIDEOLEFT,IDC_VIDEORIGHT,IDC_VIDEOMUTE,IDC_VIDEOBOTH,IDC_VIDEOLOCK,
0,0,
IDC_VIDEOACT,IDC_VIDEONAME,
"Video",
1,1,"Video",100,100,0,0,0},
{0x46,0x66,FLAG_STEREO|FLAG_MUTABLE,TYPE_DEFAULT,I90SRC_AUX,		// Aux
IDC_AUXBOX,IDC_AUXLEFT,IDC_AUXRIGHT,IDC_AUXMUTE,IDC_AUXBOTH,IDC_AUXLOCK,
0,0,
IDC_AUXACT,IDC_AUXNAME,
"Aux",
1,1,"Aux",100,100,0,0,0},
{0x4F,0x6F,FLAG_ONLYLEFT|FLAG_MUTABLE|FLAG_LOCKABLE,TYPE_DEFAULT,NO_INPUT_SEL,	// ADC
IDC_ADCBOX,IDC_ADCLEFT,IDC_ADCRIGHT,IDC_ADCMUTE,IDC_ADCBOTH,IDC_ADCLOCK,
0,0,
IDC_ADCACT,IDC_ADCNAME,
"ADC",
1,0,"ADC",100,100,0,0,0},	// ADC is marked inactive initially
{0x4D,0x6D,FLAG_STEREO|FLAG_MUTABLE|FLAG_LOCKABLE,TYPE_DEFAULT,NO_INPUT_SEL,	// DAC
IDC_DACBOX,IDC_DACLEFT,IDC_DACRIGHT,IDC_DACMUTE,IDC_DACBOTH,IDC_DACLOCK,
0,0,
IDC_DACACT,IDC_DACNAME,
"DAC",
1,1,"DAC",100,100,0,0,0},
{0x4C,0x6C,FLAG_MUTABLE,TYPE_3D,NO_INPUT_SEL,			// 3D
IDC_3DBOX,IDC_3DLEFT,IDC_3DRIGHT,IDC_3DMUTE,IDC_3DBOTH,IDC_3DLOCK,
IDC_3DLDESC,IDC_3DRDESC,
IDC_3DACT,IDC_3DNAME,
"3D",
1,1,"3D",100,100,1,0,0},
{0x4B,0x6B,FLAG_NONE,TYPE_TONE,NO_INPUT_SEL,				// Tone
IDC_TONEBOX,IDC_TONELEFT,IDC_TONERIGHT,IDC_TONEMUTE,IDC_TONEBOTH,IDC_TONELOCK,
IDC_TONELDESC,IDC_TONERDESC,
IDC_TONEACT,IDC_TONENAME,
"Tone",
1,1,"Tone",50,50,0,0,0},
{0x4E,0x6E,FLAG_LOCKABLE|FLAG_ONLYLEFT,TYPE_SELECTOR,NO_INPUT_SEL,	// Source
IDC_SELBOX,IDC_SELLEFT,IDC_SELRIGHT,IDC_SELMUTE,IDC_SELBOTH,IDC_SELLOCK,
0,0,
IDC_SELACT,IDC_SELNAME,
"Input Selector",
1,1,"Input Selector",100,0,0,0,0}
};
int NControls=12;		// Total number of controls

int	ApiLevel=0;			// ApiLevel
int	RestoreState=FALSE;		// Restore mixer settings on startup
int	AutoMini=FALSE;			// Minimize on losing focus
int	Repaint=FALSE;			/* Next DisplayControls must also
					   re-display boxes, listboxes, etc
                                           Set to TRUE at startup and after
                                           setup */
int	ChildDlg=FALSE;			// True while displaying a child window

char *InfoBuf=NULL;
size_t InfoBufSize=0;

CHAR	szTitle[MESSAGELEN];
#define MAX_APP_NAME_LEN 32
CHAR	szIniApp[MAX_APP_NAME_LEN]="LBMix";
CHAR	szSemaphore[32]="\\SEM32\\LBMIX\\";
int	UsingPipe=FALSE;
char	szPipeName[CCHMAXPATH]="\\PIPE\\MIXER";

const CHAR	szKeyWindowPos[]="WindowPos";
const CHAR	szKeyRestore[]="RestoreState";
const CHAR	szKeyAutoMini[]="AutoMinimize";
const CHAR      szLeftSuf[]="Left";
const CHAR      szRightSuf[]="Right";
const CHAR      szMuteSuf[]="Mute";
const CHAR      szBothSuf[]="Both";
const CHAR      szLockSuf[]="Lock";
const CHAR      szActiveSuf[]="Active";
const CHAR      szNameSuf[]="Name";
const CHAR      szCtlPref[]="Ctl";

int	Changing=FALSE;		// Set to TRUE while setting sliders

int PipeCommand(char *Buff,ULONG Size) {
   ULONG ulRC;
   HFILE hFile;
   ULONG A;
   ULONG rc;

   rc=DosWaitNPipe(szPipeName,PIPE_WAIT);
   if (rc) return rc;
   rc=DosOpen(szPipeName,&hFile,&A,0,0,OPEN_ACTION_OPEN_IF_EXISTS,
                OPEN_ACCESS_READWRITE|OPEN_SHARE_DENYNONE,NULL);
   if (rc) return rc;
   rc=DosWrite(hFile,Buff,strlen(Buff),&A);
   if (rc) return rc;
   rc=DosRead(hFile,Buff,Size,&A);		// This is WRONG
   if (rc) return rc;
   Buff[A]=0;
   return DosClose(hFile);
   }

int CallSet(int Fn,MIXSTRUCT *MixStruct) {
   char Buff[32];
   int rc;

   if (!UsingPipe) return mixerapiIOCTL90(Fn,MixStruct,sizeof(MIXSTRUCT));
   sprintf(Buff,"%x %d %d %x",Fn,MixStruct->VolumeL,MixStruct->VolumeR,
           MixStruct->Mute);
   rc=PipeCommand(Buff,31);
   if (rc) return rc;
// Should check for REP_OK here. Though this return code is never checked
   return 0;   
   }

int CallQuery(int Fn,MIXSTRUCT *MixStruct) {
   char Buff[32];
   int rc;

   if (!UsingPipe) return mixerapiIOCTL90(Fn,MixStruct,sizeof(MIXSTRUCT));
   _itoa(Fn,Buff,16);
   rc=PipeCommand(Buff,31);
   if (rc) return rc;
   rc=sscanf(Buff,"%x %d %d %x",&Fn,&MixStruct->VolumeL,&MixStruct->VolumeR,
             &MixStruct->Mute);
   if (rc!=4) return 1;
   return 0;
   }

void SetControl(int no) {
   MIXSTRUCT MixStruct;

   if (Controls[no].supported&&Controls[no].active) {
      MixStruct.VolumeL=Controls[no].left;
      if (Controls[no].flags&FLAG_ONLYLEFT) MixStruct.VolumeR=0;
      else MixStruct.VolumeR=Controls[no].right;
      if (Controls[no].mute) MixStruct.Mute=FLAG_MUTE;
      else MixStruct.Mute=0;
      if (Controls[no].flags&FLAG_LOCKABLE) {
         if (ApiLevel>1) {
            if (!Controls[no].lock) MixStruct.Mute|=FLAG_RELEASE;
            }
         else Controls[no].lock=TRUE;
         }
      CallSet(Controls[no].setfunc,&MixStruct);
      }
   }
void DisplayControl(int no) {
   int Enabled,EnabledL,i,j;
   HWND hwndList;

   Changing=TRUE;
   if (Repaint) WinSetWindowText(WinWindowFromID(hwndMain,Controls[no].groupID),Controls[no].name);
   Enabled=Controls[no].supported&&Controls[no].active;
   if (Controls[no].flags&FLAG_LOCKABLE) EnabledL=Enabled&&Controls[no].lock;
   else EnabledL=Enabled;
   if (Repaint) WinEnableWindow(WinWindowFromID(hwndMain,Controls[no].groupID),Enabled);

   if (!(Controls[no].flags&FLAG_ONLYRIGHT)) {
      if ((Repaint)&&(Controls[no].ldescID))
         WinEnableWindow(WinWindowFromID(hwndMain,Controls[no].ldescID),Enabled);
      WinEnableWindow(WinWindowFromID(hwndMain,Controls[no].leftID),EnabledL);
      if (Controls[no].supported) {
         if (Controls[no].type==TYPE_SELECTOR) {
            hwndList=WinWindowFromID(hwndMain,Controls[no].leftID);
            if (Repaint) {
               WinSendMsg(hwndList,LM_DELETEALL,(MPARAM)NULL,(MPARAM)NULL );
               for (i=0;i<NControls;i++) if ((Controls[i].sel!=NO_INPUT_SEL)&&(Controls[i].active)&&(Controls[i].supported))
                  WinSendMsg(hwndList,LM_INSERTITEM,MPFROMSHORT(LIT_END),
                             MPFROMP((PSZ)Controls[i].name));
               }
            for (i=0,j=0;i<NControls;i++) if ((Controls[i].sel!=NO_INPUT_SEL)&&(Controls[i].active)&&(Controls[i].supported)) {
               if (Controls[i].sel&Controls[no].left)
                  WinSendMsg(hwndList,LM_SELECTITEM,(MPARAM)j,(MPARAM)TRUE);
               else
                  WinSendMsg(hwndList,LM_SELECTITEM,(MPARAM)j,(MPARAM)FALSE);
               j++;
               }
            }
         else WinSendDlgItemMsg(hwndMain,Controls[no].leftID,
               SLM_SETSLIDERINFO,MPFROM2SHORT(SMA_SLIDERARMPOSITION,
               SMA_INCREMENTVALUE), (MPARAM)Controls[no].left);
         }
      }
   if (!(Controls[no].flags&FLAG_ONLYLEFT)) {
      if ((Repaint)&&(Controls[no].rdescID))
         WinEnableWindow(WinWindowFromID(hwndMain,Controls[no].rdescID),Enabled);
      WinEnableWindow(WinWindowFromID(hwndMain,Controls[no].rightID),EnabledL);
      if (Controls[no].supported) WinSendDlgItemMsg(hwndMain,Controls[no].rightID,
            SLM_SETSLIDERINFO,MPFROM2SHORT(SMA_SLIDERARMPOSITION,
            SMA_INCREMENTVALUE), (MPARAM)Controls[no].right);
      }
   if (Controls[no].flags&FLAG_MUTABLE) {
      WinEnableWindow(WinWindowFromID(hwndMain,Controls[no].muteID),EnabledL);
      if (Controls[no].supported) WinSendDlgItemMsg(hwndMain,Controls[no].muteID,BM_SETCHECK,
            (MPARAM)Controls[no].mute,0);
      }
   if (Controls[no].flags&FLAG_STEREO) {
      WinEnableWindow(WinWindowFromID(hwndMain,Controls[no].bothID),Enabled);
      if (Controls[no].supported) WinSendDlgItemMsg(hwndMain, Controls[no].bothID,BM_SETCHECK,
            (MPARAM)Controls[no].both,0);
      }
   if (Controls[no].flags&FLAG_LOCKABLE) {
      if (ApiLevel>1)
         WinEnableWindow(WinWindowFromID(hwndMain,Controls[no].lockID),Enabled);
      else
         WinEnableWindow(WinWindowFromID(hwndMain,Controls[no].lockID),0);
      if (Controls[no].supported) WinSendDlgItemMsg(hwndMain,Controls[no].lockID,BM_SETCHECK,
            (MPARAM)Controls[no].lock,0);
      }
   Changing=FALSE;
   return;
   }

void QueryControl(int no) {
   MIXSTRUCT MixStruct;

   if (Controls[no].supported)
      if (!CallQuery(Controls[no].qfunc,&MixStruct)) {
         Controls[no].left=MixStruct.VolumeL;
         Controls[no].right=MixStruct.VolumeR;
         Controls[no].mute=(MixStruct.Mute&FLAG_MUTE);
         if ((Controls[no].flags&FLAG_LOCKABLE)&&(ApiLevel>1)) 
            Controls[no].lock=!(MixStruct.Mute&FLAG_RELEASE);
         }
   return;
   }

void SetControls() {
   int i;

   for (i=0;i<NControls;i++) SetControl(i);
   return;
   }
void DisplayControls() {
   int i;

   for (i=0;i<NControls;i++) DisplayControl(i);
   Repaint=FALSE;
   return;
   }

void QueryControls() {
   int i;

   for (i=0;i<NControls;i++) QueryControl(i);
   return;
   }

void InitControls() {
   int i;

   for (i=0;i<NControls;i++) if ((Controls[i].supported)&&
                                  (Controls[i].type==TYPE_SELECTOR)&&
                                  (Controls[i].flags&FLAG_MULTIPLE))
       WinSetWindowBits(WinWindowFromID(hwndMain,Controls[i].leftID),
                        QWL_STYLE,LS_MULTIPLESEL|WS_DISABLED,
                        LS_MULTIPLESEL|WS_DISABLED);
   }

// Open device, query available controls and set up Controls structure
// return TRUE if Ok.
int ApiInit() {
   unsigned char ApiMap[256];
   int i;
   MIXSTRUCT MixStruct;
   MIXMSGBUF MixMsgBuf;
   ULONG Level;
   char Buff[PIPE_BUF_MAX+1];
   char *tp;

   if (!UsingPipe) {
      if (mixerapiInit(MyCallBack)) UsingPipe=TRUE;	// Try pipe API
      else {
         if (!mixerapiIOCTL90(GETAPIMAP,ApiMap,256)) {
            for (i=0;i<NControls;i++)
               if (!(ApiMap[Controls[i].setfunc]&&ApiMap[Controls[i].qfunc])) {
                  Controls[i].supported=0;
                  Controls[i].active=0;
                  }
            }
         if (!mixerapiIOCTL90(APILEVELQUERY,&Level,sizeof(ULONG))) ApiLevel=Level;
         }
      }

   if (UsingPipe) {
      strcpy(Buff,"80");		// Api Level
      if (PipeCommand(Buff,PIPE_BUF_MAX)) return 0;
      ApiLevel=atoi(Buff+3);
      }

// Check if query functions work
   for (i=0;i<NControls;i++) if (Controls[i].supported) {
       if (CallQuery(Controls[i].qfunc,&MixStruct)) {
            Controls[i].supported=0;
            Controls[i].active=0;
            }
       else {
            if (Controls[i].type==TYPE_SELECTOR) {
               if (MixStruct.Mute&4) Controls[i].flags|=FLAG_MULTIPLE;
               }
            else if (Controls[i].type==TYPE_TONE) {
               if (!(MixStruct.Mute&1)) Controls[i].flags|=FLAG_ONLYRIGHT;
               if (!(MixStruct.Mute&2)) Controls[i].flags|=FLAG_ONLYLEFT;
               }
            else if (Controls[i].type==TYPE_3D) {
               if (!(MixStruct.Mute&2)) Controls[i].flags|=FLAG_ONLYRIGHT;
               if (!(MixStruct.Mute&4)) Controls[i].flags|=FLAG_ONLYLEFT;
               }
            }
       }
   for (i=0;i<NControls;i++) if (Controls[i].supported) break; 
   if (i==NControls) return 0;	// No controls are available 
// If API Level >1 then default for lockables is "active"
   if (ApiLevel>1) for (i=0;i<<NControls;i++) {
      if ((Controls[i].supported)&&(Controls[i].flags&FLAG_LOCKABLE))
         Controls[i].active=TRUE;
      }

   if (!UsingPipe) {
      if (ApiMap[MSGBUF]) {
         MixMsgBuf.pBuffer=0;
         MixMsgBuf.ulSize=0;
         MixMsgBuf.fClear=FALSE;
         if (!mixerapiIOCTL90(MSGBUF,&MixMsgBuf,sizeof(MIXMSGBUF))) {
            InfoBufSize=MixMsgBuf.ulSize;
            InfoBuf=malloc(InfoBufSize);
            MixMsgBuf.pBuffer=(ULONG)InfoBuf;
            MixMsgBuf.ulSize=InfoBufSize;
            MixMsgBuf.fClear=FALSE;
            if (mixerapiIOCTL90(MSGBUF,&MixMsgBuf,sizeof(MIXMSGBUF))) {
               free(InfoBuf);
               InfoBuf=NULL;
               }
            }
         }
      }
   else { 
      strcpy(Buff,"83");
      if (!PipeCommand(Buff,PIPE_BUF_MAX)) {
         tp=strchr(Buff,' ');
         if (tp) {
            if (*tp) tp++;
            if (*tp) {
               InfoBufSize=strlen(tp)+1;
               InfoBuf=malloc(InfoBufSize);
               strcpy(InfoBuf,tp);
               }
            }
         }
      }

   if (!InfoBuf) {
      InfoBufSize=DEF_INFO_BUF_SIZE;
      InfoBuf=malloc(InfoBufSize);
      strcpy(InfoBuf,"Driver does not return info.");
      strcat(InfoBuf,"\nAPI Level: ");
      _itoa(ApiLevel,InfoBuf+strlen(InfoBuf),10);
      strcat(InfoBuf,"\nMixer state change callback is ");
      if (!ApiMap[CALLBACKREG]) strcat(InfoBuf,"not ");
      strcat(InfoBuf,"supported.");
      strcat(InfoBuf,"\nSupported controls:");
      for (i=0;i<NControls;i++) if (Controls[i].supported) {
         strcat(InfoBuf,"\n");
         strcat(InfoBuf,Controls[i].defname);
         }
      }

   return 1;
   }

void ApiClose() {
   if (!UsingPipe) mixerapiDeInit();
   }

void LoadProfile() {
   unsigned char tname[32];
   char Buff[32];
   int i;
   unsigned char *SuffPos;

   if (PrfQueryProfileString(HINI_USERPROFILE,(char *)szIniApp,(char *)szKeyRestore,NULL,Buff,32))
      RestoreState=atoi(Buff);
   if (PrfQueryProfileString(HINI_USERPROFILE,(char *)szIniApp,(char *)szKeyAutoMini,NULL,Buff,32))
      AutoMini=atoi(Buff);

   for (i=0;i<NControls;i++) {
      strcpy(tname,szCtlPref);
      _itoa(Controls[i].setfunc,tname+strlen(tname),16);
      SuffPos=tname+strlen(tname);

      strcpy(SuffPos,szLeftSuf);
      if (PrfQueryProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,NULL,Buff,32))
         Controls[i].left=atoi(Buff);
      strcpy(SuffPos,szRightSuf);
      if (PrfQueryProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,NULL,Buff,32))
         Controls[i].right=atoi(Buff);
      strcpy(SuffPos,szMuteSuf);
      if (PrfQueryProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,NULL,Buff,32))
         Controls[i].mute=atoi(Buff);
      strcpy(SuffPos,szBothSuf);
      if (PrfQueryProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,NULL,Buff,32))
         Controls[i].both=atoi(Buff);
      if (ApiLevel>1) {
         strcpy(SuffPos,szLockSuf);
         if (PrfQueryProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,NULL,Buff,32))
            Controls[i].lock=atoi(Buff);
         }
      strcpy(SuffPos,szActiveSuf);
      if (PrfQueryProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,NULL,Buff,32))
         Controls[i].active=atoi(Buff);
      strcpy(SuffPos,szNameSuf);
      PrfQueryProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,NULL,Controls[i].name,32);
      }
   return;
   }

void SaveProfile() {
   unsigned char tname[32];
   char Buff[32];
   int i;
   unsigned char *SuffPos;

   _itoa(RestoreState,Buff,10);
   PrfWriteProfileString(HINI_USERPROFILE,(char *)szIniApp,(char *)szKeyRestore,Buff);
   _itoa(AutoMini,Buff,10);
   PrfWriteProfileString(HINI_USERPROFILE,(char *)szIniApp,(char *)szKeyAutoMini,Buff);

   for (i=0;i<NControls;i++) {
      strcpy(tname,szCtlPref);
      _itoa(Controls[i].setfunc,tname+strlen(tname),16);
      SuffPos=tname+strlen(tname);

      strcpy(SuffPos,szLeftSuf);
      _itoa(Controls[i].left,Buff,10);
      PrfWriteProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,Buff);
      strcpy(SuffPos,szRightSuf);
      _itoa(Controls[i].right,Buff,10);
      PrfWriteProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,Buff);
      strcpy(SuffPos,szMuteSuf);
      _itoa(Controls[i].mute,Buff,10);
      PrfWriteProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,Buff);
      strcpy(SuffPos,szBothSuf);
      _itoa(Controls[i].both,Buff,10);
      PrfWriteProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,Buff);
      strcpy(SuffPos,szLockSuf);
      _itoa(Controls[i].lock,Buff,10);
      PrfWriteProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,Buff);
      strcpy(SuffPos,szActiveSuf);
      _itoa(Controls[i].active,Buff,10);
      PrfWriteProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,Buff);
      strcpy(SuffPos,szNameSuf);
      PrfWriteProfileString(HINI_USERPROFILE,(char *)szIniApp,tname,Controls[i].name);
      }
   return;
   }

int ParseParms(int argc, char *argv[]) {
   int i;

   for (i=1;i<argc;i++) {
     if (!stricmp(argv[i],"-i"))  {
        i++;
        if (i==argc) return FALSE;		// Syntax error
        strncpy(szIniApp,argv[i],MAX_APP_NAME_LEN-1);
        }
     else if (!stricmp(argv[i],"-d"))  {
        i++;
        if (i==argc) return FALSE;		// Syntax error
        strcpy(szPddName,"\\DEV\\");
        strncpy(szPddName+5,argv[i],8);
        strupr(szPddName);
        }
     else if (!stricmp(argv[i],"-p"))
        UsingPipe=TRUE;
     else if (!stricmp(argv[i],"-pn"))  {
        i++;
        if (i==argc) return FALSE;		// Syntax error
        UsingPipe=TRUE;
        strncpy(szPipeName,argv[i],CCHMAXPATH-1);
        }
     else return FALSE;				// Syntax error
     }
   return TRUE;
   }

INT main(int argc, char *argv[])
{
   QMSG qmsg;					      /* message structure */

   hab = WinInitialize(0);
   if(!hab)  {
       DosBeep(BEEP_WARN_FREQ, BEEP_WARN_DUR);
       return(RETURN_ERROR);
   }
   hmq = WinCreateMsgQueue(hab, 0);
   if(!hmq)  {
       DosBeep(BEEP_WARN_FREQ, BEEP_WARN_DUR);
       WinTerminate(hab);
       return(RETURN_ERROR);
   }
   if(!Init()) {
       MessageBox(HWND_DESKTOP, IDMSG_INITFAILED, "Error !",
			   MB_OK | MB_ERROR | MB_MOVEABLE, TRUE);
       DosExit(EXIT_PROCESS, RETURN_ERROR);
       }
   if(!ParseParms(argc,argv)) {
       MessageBox(HWND_DESKTOP, IDMSG_SYNTAXERROR, "Error !",
			   MB_OK | MB_ERROR | MB_MOVEABLE, TRUE);
       DosExit(EXIT_PROCESS, RETURN_ERROR);
       }
   if(!ApiInit()) {
       MessageBox(HWND_DESKTOP, IDMSG_APIINITFAILED, "Error !",
			   MB_OK | MB_ERROR | MB_MOVEABLE, TRUE);
       DosExit(EXIT_PROCESS, RETURN_ERROR);
       }
   if ((ApiLevel<=1)&&(!UsingPipe)) {
      strcat(szSemaphore,szPddName+5);
      if (DosCreateEventSem((char *)szSemaphore,&hev,DC_SEM_SHARED,FALSE)) {
         MessageBox(HWND_DESKTOP, IDMSG_ALREADYRUNNING, "Error !",
			   MB_OK | MB_ERROR | MB_MOVEABLE, TRUE);
         ApiClose();
         DosExit(EXIT_PROCESS, RETURN_ERROR);
         }
      }
   InitControls();

/*   InitHelp(); */

   while(WinGetMsg(hmq, (PQMSG)&qmsg, 0L, 0L, 0L))
	   WinDispatchMsg(hmq, (PQMSG)&qmsg);
					   /* destroy the help instance */
/*   DestroyHelpInstance(); */
   ApiClose();
   if (hev) DosCloseEventSem(hev);
   if (InfoBuf) free(InfoBuf);
				    /* will normally be put in ExitProc */
   DosExit(EXIT_PROCESS, RETURN_SUCCESS);
   return 0;
}							       /* main() */

BOOL Init(VOID)
{
    SWCNTRL SwData;
    PID pid;

    /* Add ExitProc to the exit list to handle the exit processing */
    if(DosExitList(EXLST_ADD, ExitProc)) return FALSE;

   WinLoadString(hab, 0, IDS_TITLE, MESSAGELEN, szTitle);



/* Create subwindows */
    hwndMain=WinLoadDlg(HWND_DESKTOP, HWND_DESKTOP, (PFNWP)MainDlgProc,
			(HMODULE) NULL, ID_RESOURCE, NULL);
    WinSetWindowText(hwndMain,szTitle);
    WinQueryWindowProcess(hwndMain,&pid,NULL);
    SwData.hwnd=hwndMain;
    SwData.hwndIcon=NULLHANDLE;
    SwData.hprog=NULLHANDLE;
    SwData.idProcess=pid;
    SwData.idSession=0;
    SwData.uchVisibility=SWL_VISIBLE;
    SwData.fbJump=SWL_JUMPABLE;
    WinLoadString(hab, 0, IDS_SWTITLE, MESSAGELEN, SwData.szSwtitle);
    hswitch=WinCreateSwitchEntry(hab,&SwData);
    return TRUE;
}							  /* Init() */

VOID APIENTRY ExitProc(ULONG usTermCode)
				  /* code for the reason for termination */
{
   WinDestroyWindow(hwndMain);
   WinDestroyMsgQueue(hmq);
   WinTerminate(hab);
   DosExitList(EXLST_EXIT, (PFNEXITLIST)0L);	/* termination complete */

   return;
}							   /* ExitProc() */
LONG MessageBox(HWND hwndOwner, LONG IdMsg, PSZ pszMsg, LONG fsStyle,
		     BOOL bBeep)
{
    CHAR szText[MESSAGELEN];
    LONG usRet;

    if(!WinLoadMessage(hab, (HMODULE)NULL, IdMsg, MESSAGELEN, (PSZ)szText)) {
	WinAlarm(HWND_DESKTOP, WA_ERROR);
	return RETURN_ERROR;
        }
    if(bBeep) WinAlarm(HWND_DESKTOP, WA_ERROR);

    usRet = WinMessageBox(HWND_DESKTOP,
			 hwndOwner,
			 szText,
			 (PSZ)pszMsg,
			 IDM_MSGBOX,
			 fsStyle);
    return usRet;
}						    /* MessageBox() */


void MainControl(HWND hwnd,MPARAM mp1) {
   int i,hit,j,k,val;

   if ((!Changing)&&((SHORT2FROMMP(mp1)==BN_CLICKED)||
       (SHORT2FROMMP(mp1)==LN_SELECT)||(SHORT2FROMMP(mp1)==SLN_CHANGE)||
       (SHORT2FROMMP(mp1)==SLN_SLIDERTRACK))) {
      for (i=0;i<NControls;i++) if(Controls[i].supported&Controls[i].active) {
         hit=TRUE;
         if (LOUSHORT(mp1)==Controls[i].leftID) {
            if (Controls[i].type==TYPE_SELECTOR) {
               Controls[i].left=0;
               val=LIT_FIRST;
               while(1) {
                  val=(int)WinSendDlgItemMsg(hwnd,Controls[i].leftID,
                       LM_QUERYSELECTION,(MPARAM)val,(MPARAM)NULL);
                  if (val==LIT_NONE) break;
                  for (k=0,j=0;k<NControls;k++) if ((Controls[k].sel!=NO_INPUT_SEL)&&(Controls[k].active)&&(Controls[k].supported)) {
                     if (j==val) {
                        Controls[i].left|=Controls[k].sel;
                        break;
                        }
                     else j++;
                     }
                 }
//               if (Controls[i].flags&FLAG_LOCKABLE) Controls[i].lock=TRUE;
               }
            else {
               Controls[i].left=(int)WinSendDlgItemMsg(hwnd,Controls[i].leftID,
                    SLM_QUERYSLIDERINFO, MPFROM2SHORT(SMA_SLIDERARMPOSITION,
                    SMA_INCREMENTVALUE),(MPARAM)NULL);
               if ((Controls[i].flags&FLAG_STEREO)&&Controls[i].both)
                   Controls[i].right=Controls[i].left;
//               if (Controls[i].flags&FLAG_LOCKABLE) Controls[i].lock=TRUE;
               }
            }
         else if (LOUSHORT(mp1)==Controls[i].rightID) {
            if (!(Controls[i].flags&FLAG_ONLYLEFT)) {
               Controls[i].right=(int)WinSendDlgItemMsg(hwnd,Controls[i].rightID,
                    SLM_QUERYSLIDERINFO, MPFROM2SHORT(SMA_SLIDERARMPOSITION,
                    SMA_INCREMENTVALUE),(MPARAM)NULL);
               if ((Controls[i].flags&FLAG_STEREO)&&Controls[i].both)
                   Controls[i].left=Controls[i].right;
//               if (Controls[i].flags&FLAG_LOCKABLE) Controls[i].lock=TRUE;
               }
            }
         else if (LOUSHORT(mp1)==Controls[i].muteID) {
            if (Controls[i].flags&FLAG_MUTABLE) {
               Controls[i].mute=(int)WinSendDlgItemMsg(hwnd, Controls[i].muteID,
                                                       BM_QUERYCHECK,0,0);
//               if (Controls[i].flags&FLAG_LOCKABLE) Controls[i].lock=TRUE;
               }
            }
         else if (LOUSHORT(mp1)==Controls[i].bothID) {
            if (Controls[i].flags&FLAG_STEREO) {
               Controls[i].both=(int)WinSendDlgItemMsg(hwnd, Controls[i].bothID,
                                                       BM_QUERYCHECK,0,0);
               if ((Controls[i].both)&&(Controls[i].right!=Controls[i].left)) {
                   Controls[i].left=(Controls[i].left+Controls[i].right)/2;
                   Controls[i].right=Controls[i].left;
                   }
               }
            }
         else if (LOUSHORT(mp1)==Controls[i].lockID) {
            if (Controls[i].flags&FLAG_LOCKABLE)
               Controls[i].lock=(int)WinSendDlgItemMsg(hwnd, Controls[i].lockID,
                                                       BM_QUERYCHECK,0,0);
            }
         else hit=FALSE;
         if (hit) {
            SetControl(i);
            QueryControl(i);
            DisplayControl(i);
            break;
            }
         }
      }
   return;
   }

void MainCommand(HWND hwnd,MPARAM mp1) {

   switch(LOUSHORT(mp1)) {
      case IDC_SETTINGS:
         ChildDlg=TRUE;
         WinDlgBox(HWND_DESKTOP, hwndMain, (PFNWP)SettingsDlgProc,
 			(HMODULE) NULL, IDD_SETTINGS, NULL);
         Repaint=TRUE;
         ChildDlg=FALSE;
         DisplayControls();
         break;
      case IDC_ABOUT:
         ChildDlg=TRUE;
         WinDlgBox(HWND_DESKTOP, hwndMain, (PFNWP)AboutDlgProc,
			(HMODULE) NULL, IDD_ABOUT, NULL);
         ChildDlg=FALSE;
         break;
      case IDC_INFO:
         ChildDlg=TRUE;
         WinDlgBox(HWND_DESKTOP, hwndMain, (PFNWP)InfoDlgProc,
			(HMODULE) NULL, IDD_INFO, NULL);
         ChildDlg=FALSE;
         break;
      case IDC_EXIT:
         WinPostMsg( hwndMain, WM_CLOSE, (MPARAM)0,(MPARAM)0 );
         break;
      }
   return;
   }



MRESULT EXPENTRY MainDlgProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   MRESULT sRC;
   SWP swpCurrent;

     switch (msg) {
       case WM_INITDLG:
          WinRestoreWindowPos((char *)szIniApp,(char *)szKeyWindowPos,hwnd);
          WinSendMsg(hwnd,WM_SETICON,
            (MPARAM) WinLoadPointer(HWND_DESKTOP,(HMODULE) NULL,ID_RESOURCE ),
            (MPARAM) 0 );
          WinPostMsg( hwnd, WM_USER, (MPARAM)USER_INIT,(MPARAM)0 );
 	  break;
       case WM_USER:
          if ((int)mp1==USER_INIT) {
             LoadProfile();
             if (!RestoreState) QueryControls();
             else SetControls();
             Repaint=TRUE;
             DisplayControls();
             }
          else if ((int)mp1==USER_QUERY) {
             QueryControls();
             DisplayControls();
             }
          break;
       case WM_CLOSE:
          ChildDlg=TRUE;	// To prevent temporary minimization
          SaveProfile();
          WinStoreWindowPos((char *)szIniApp,(char *)szKeyWindowPos,hwnd);
          WinPostMsg( hwndMain, WM_QUIT, (MPARAM)0,(MPARAM)0 );
          break;
       case WM_COMMAND:
          MainCommand(hwnd,mp1);
          break;
       case WM_CONTROL:
          MainControl(hwnd,mp1);
          break;
       case WM_ACTIVATE:
          if ((!SHORT1FROMMP(mp1))&&(AutoMini)&&(!ChildDlg)) {
             WinQueryWindowPos(hwnd,&swpCurrent);
             WinSetWindowPos(hwnd,0,swpCurrent.x,swpCurrent.y,
                             swpCurrent.cx,swpCurrent.cy,
                             SWP_MINIMIZE|SWP_SIZE|SWP_MOVE);
             }
       default:	
	  sRC = WinDefDlgProc(hwnd, msg, mp1, mp2);
	  return sRC;
       }
   return (MRESULT)0L;
}

int MyCallBack() {
   WinPostMsg( hwndMain, WM_USER, (MPARAM)USER_QUERY,(MPARAM)0 );
   return;
   }


void SettingsInit(HWND hwnd) {
   int i;

   WinSendDlgItemMsg(hwnd, IDC_RESTORE,BM_SETCHECK,(MPARAM)RestoreState,0);
   WinSendDlgItemMsg(hwnd, IDC_AUTOMINI,BM_SETCHECK,(MPARAM)AutoMini,0);
   for(i=0;i<NControls;i++) {
      WinSendDlgItemMsg(hwnd,Controls[i].activeID,BM_SETCHECK,(MPARAM)Controls[i].active,0);
      WinSetWindowText(WinWindowFromID(hwnd,Controls[i].nameID),Controls[i].name);
      if (Controls[i].supported) {
         WinEnableWindow(WinWindowFromID(hwnd,Controls[i].activeID),TRUE);
         WinEnableWindow(WinWindowFromID(hwnd,Controls[i].nameID),TRUE);
         }
      }
   return;
   }

void SettingsCommand(HWND hwnd,MPARAM mp1) {
   int i;
   char Buff[32];
   unsigned int OldBase;

   switch(LOUSHORT(mp1)) {
      case IDC_CANCEL:
         WinDismissDlg(hwnd, FALSE);
         break;
      case IDC_DEFAULTS:
         WinSendDlgItemMsg(hwnd, IDC_RESTORE,BM_SETCHECK,(MPARAM)FALSE,0);
         WinSendDlgItemMsg(hwnd, IDC_AUTOMINI,BM_SETCHECK,(MPARAM)FALSE,0);
         for (i=0;i<NControls;i++) {
            if ((ApiLevel>1)||(!(Controls[i].flags&FLAG_LOCKABLE))) WinSendDlgItemMsg(hwnd,Controls[i].activeID,BM_SETCHECK,(MPARAM)TRUE,0);
            else WinSendDlgItemMsg(hwnd,Controls[i].activeID,BM_SETCHECK,(MPARAM)TRUE,0);
            WinSetWindowText(WinWindowFromID(hwnd,Controls[i].nameID),Controls[i].defname);
            } 
         break;
      case IDC_OK:
         RestoreState=(int)WinSendDlgItemMsg(hwnd,IDC_RESTORE,BM_QUERYCHECK,0,0);
         AutoMini=(int)WinSendDlgItemMsg(hwnd,IDC_AUTOMINI,BM_QUERYCHECK,0,0);
         for(i=0;i<NControls;i++) if (Controls[i].supported) {
            Controls[i].active=(int)WinSendDlgItemMsg(hwnd,Controls[i].activeID,BM_QUERYCHECK,0,0);
            WinQueryWindowText(WinWindowFromID(hwnd,Controls[i].nameID),32,Controls[i].name);
            }
         SaveProfile();
         WinDismissDlg(hwnd, FALSE);
         break;
      }
   return;
   }

MRESULT EXPENTRY SettingsDlgProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   MRESULT sRC;

     switch (msg) {
       case WM_INITDLG:
          SettingsInit(hwnd);
 	  break;
       case WM_CLOSE:
          WinDismissDlg(hwnd, FALSE);
          break;
       case WM_COMMAND:
          SettingsCommand(hwnd,mp1);
          break;
/*       case WM_CONTROL:
          SettingsControl(hwnd,mp1);
          break; */
       default:	
	  sRC = WinDefDlgProc(hwnd, msg, mp1, mp2);
	  return sRC;
       }
   return (MRESULT)0L;
}

MRESULT EXPENTRY AboutDlgProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   MRESULT sRC;

     switch (msg) {
       case WM_INITDLG:
          WinSetWindowText(WinWindowFromID(hwnd,IDC_ABOUTAPP),szTitle);
 	  break;
       case WM_COMMAND:
          if (LOUSHORT(mp1)==IDC_OK) WinDismissDlg(hwnd, TRUE);
          break;
       default:	
	  sRC = WinDefDlgProc(hwnd, msg, mp1, mp2);
	  return sRC;
       }
   return (MRESULT)0L;
}

MRESULT EXPENTRY InfoDlgProc(HWND hwnd, ULONG msg, MPARAM mp1, MPARAM mp2)
{
   MRESULT sRC;
   IPT iptOffset;
   HWND hwndMLE;

     switch (msg) {
       case WM_INITDLG:
          {
          char tbuf[80];
          char tbuf1[10];

          hwndMLE=WinWindowFromID(hwnd,IDC_INFOMLE);
          WinSendMsg(hwndMLE, MLM_DISABLEREFRESH, NULL, NULL);
          WinSendMsg(hwndMLE, MLM_SETSEL, MPFROMSHORT(NULL),
                     (MPARAM) WinSendMsg(hwndMLE, MLM_QUERYTEXTLENGTH,NULL, NULL));
          WinSendMsg(hwndMLE, MLM_CLEAR, NULL, NULL);
          WinSendMsg(hwndMLE, MLM_SETIMPORTEXPORT,
                       MPFROMP((PBYTE)InfoBuf),(MPARAM)strlen(InfoBuf));
          iptOffset=0;
          WinSendMsg(hwndMLE, MLM_IMPORT, MPFROMP(&iptOffset),
                          (MPARAM)strlen(InfoBuf));
          WinSendMsg(hwndMLE, MLM_ENABLEREFRESH, NULL, NULL);
          if (UsingPipe) {
             WinLoadString(hab, 0, IDS_USINGPIPE, 79, tbuf);
             strncat(tbuf,szPipeName,79);
             }
          else {
             WinLoadString(hab, 0, IDS_USINGDRV, 79, tbuf);
             strncat(tbuf,szPddName+5,79);
             }
          if (strlen(tbuf)<50) {
             WinLoadString(hab, 0, IDS_APILEVEL, 79-strlen(tbuf), tbuf+strlen(tbuf));
             _itoa(ApiLevel,tbuf1,10);
             strncat(tbuf,tbuf1,79);
             }
          WinSetWindowText(WinWindowFromID(hwnd,IDC_INFODRV),tbuf);
 	  break;
          }
       case WM_COMMAND:
          if (LOUSHORT(mp1)==IDC_OK) WinDismissDlg(hwnd, TRUE);
          break;
       default:	
	  sRC = WinDefDlgProc(hwnd, msg, mp1, mp2);
	  return sRC;
       }
   return (MRESULT)0L;
}
