/* 

Pipe Mixer for IOCTL90 API by Lesha Bogdanow
Copyright (C) 1999-2000  Lesha Bogdanow

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#define USE_OS2_TOOLKIT_HEADERS
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <os2.h>

#include "ioctl90.h"
#include "mixerapi.h"
#include "PipeMixPvt.h"

#define DEF_INFO_BUF_SIZE 128

const char szLogo[]="PipeMixer demon for IOCTL90 API. v.0.01";

unsigned char ApiMap[256];

// Initialize mixer. Return TRUE if Ok
int MixerInit(int argc, char *argv[]) {
   int DriverInfo;
   MIXSTRUCT dummymix;
   MIXMSGBUF MixMsgBuf;
   ULONG Level;
   int i;


   for (i=1;i<argc;i++) {
     if (!stricmp(argv[i],"-d"))  {
        i++;
        if (i==argc) return FALSE;		// Syntax error
        strcpy(szPddName,"\\DEV\\");
        strncpy(szPddName+5,argv[i],8);
        strupr(szPddName);
        }
     else return FALSE;				// Syntax error
     }

   if (mixerapiInit(NULL)) return FALSE;
   if (mixerapiIOCTL90(GETAPIMAP,ApiMap,256)) return FALSE;
   if (!mixerapiIOCTL90(APILEVELQUERY,&Level,sizeof(ULONG))) ApiLevel=Level;
   if (ApiMap[MSGBUF]) {
      MixMsgBuf.pBuffer=0;
      MixMsgBuf.ulSize=0;
      MixMsgBuf.fClear=FALSE;
      if (!mixerapiIOCTL90(MSGBUF,&MixMsgBuf,sizeof(MIXMSGBUF))) {
         InfoBufSize=MixMsgBuf.ulSize;
         InfoBuf=malloc(InfoBufSize);
         MixMsgBuf.pBuffer=(ULONG)InfoBuf;
         MixMsgBuf.ulSize=InfoBufSize;
         MixMsgBuf.fClear=FALSE;
         if (!mixerapiIOCTL90(MSGBUF,&MixMsgBuf,sizeof(MIXMSGBUF))) {
            DriverInfo=TRUE;
            }
         else free(InfoBuf);
         }
      }
   if (!DriverInfo) {
      InfoBufSize=DEF_INFO_BUF_SIZE;
      InfoBuf=malloc(InfoBufSize);
      strcpy(InfoBuf,"Driver does not return info.");
      strcat(InfoBuf,"\nAPI Level: ");
      _itoa(ApiLevel,InfoBuf+strlen(InfoBuf),10);
      }
   if (Debug) printf("Device: %s, API Level: %d\nDriver Info: \n%s\n",szPddName,
                     ApiLevel,InfoBuf);
   return TRUE;
   }

// Deinitialize mixer
void MixerClose() {
   mixerapiDeInit();
   }

/* Process a command. Buff is an output buffer, the function number and
   a trailing space are already there. tokens - number of command tokens
   cmd - command, p1, p2, p3 - parameters (if tokens>= 2, 3, 4 respectively)
*/
int ProcessCmd(char *Buff, int tokens, int cmd, int p1, int p2, int p3) {
   MIXSTRUCT MixStruct;

   if (ApiMap[cmd]&&((cmd&0xF0)==0x40)) {	// Set command
      if (tokens<2) p1=100;
      if (tokens<3) p2=p1;
      if (tokens<4) p3=0;
      MixStruct.VolumeL=p1;
      MixStruct.VolumeR=p2;
      MixStruct.Mute=p3;
      if (!mixerapiIOCTL90(cmd,&MixStruct,sizeof(MIXSTRUCT))) {
         strcat(Buff,REP_OK);
         }
      else strcat(Buff,REP_FAILED);
      }
   else if (ApiMap[cmd]&&((cmd&0xF0)==0x60)) {	// Query command
      if (!mixerapiIOCTL90(cmd,&MixStruct,sizeof(MIXSTRUCT))) {
         sprintf(Buff+strlen(Buff),"%d %d %x",MixStruct.VolumeL,MixStruct.VolumeR,MixStruct.Mute);
         }
      else strcat(Buff,REP_FAILED);
      }
   else strcat(Buff,REP_UNSUPPORTED);
   return TRUE;
   }
